/****************************************************************************
                                                                   
         DT2851 LIBRARY DEMONSTRATION PROGRAM - UNIX VERSION       
 Copyright (C) 1992...1997 Walter Fetter Lages <w.fetter@ieee.org> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
                                                                   
****************************************************************************/


#include <curses.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/mman.h>
#include <unistd.h>

#include <dt2851.h>

#define xor(a,b) ((~a & b)|(a & ~b))

#define BOARD 		0

static volatile int syncflag=0;
static volatile int doneflag=0;

static int kbhit(void)
{
	int c;
	
	timeout(0);
	c=getch();
	timeout(-1);
	if(c==ERR) return 0;
	ungetch(c);
	return 1;
}

static void pausemsg(void)
{
	int x;
	int y;

	getsyx(y,x);
	move(22,0);
	printw("Press any key to continue...\n");
	move(y,x);
	refresh();
}

static void pause51(void)
{
	pausemsg();
	getch();
}

static void inttest(int dummy)
{
	if(!busy51(BOARD))
	{
		doneflag=1;
		start51(BOARD,SINGLE51);
	}
	if(vsync51(BOARD)) syncflag=1;
	intack51(BOARD);
}

static void showstatus(void)
{
	int x;
	int y;

	move(2,0);
	printw("DT2851 current status:\n\n");
	printw("Mode: %d ",getmode51(BOARD));
	switch(getmode51(BOARD))
	{
		case FEEDBACK51:
		{
			printw("feedback\n");
			break;
		}
		case VIDEO_IN51:
		{
			printw("video input\n");
			break;
		}
		case LOAD_LUT51:
		{
			printw("load look-up table\n");
			break;
		}
		case SLOW_SCAN51:
		{
			printw("slow scan input\n");
			break;
		}
		case PORT_OUT51:
		{
			printw("port output\n");
			break;
		}
		case PORT_IN51:
		{
			printw("port input\n");
			break;
		}
		default:
		{
			printw("unknown mode\n");
			break;
		}
	}
	printw("Input look-up table: %d\n",inlut51(BOARD));
	printw("Output look-up table: %d\n",outlut51(BOARD));
	printw("Input buffer: %d\n",inbuf51(BOARD));
	printw("Output buffer: %d\n",outbuf51(BOARD));
	printw("Protected planes code: %d, planes: ",protected51(BOARD));
	if(protected51(BOARD) & PLANE_7_6_5_4_51) printw("7 6 5 4 ");
	if(protected51(BOARD) & PLANE_3_2_51) printw("3 2 ");
	if(protected51(BOARD) & PLANE_1_51) printw("1 ");
	if(protected51(BOARD) & PLANE_0_51) printw("0");
	if(!protected51(BOARD)) printw("no one");
	printw("\n");
	if(busy51(BOARD)) printw("Board is busy\t\n"); else printw("Board is not busy\n");
	if(cursor51(BOARD)) printw("Cursor on\t\n"); else printw("Cursor off\n");
	getcursorpos51(BOARD,&x,&y);
	printw("Cursor at: %d, %d\n",x,y);
	if(timing51(BOARD) & EXTTIM51) printw("External timing\n"); else printw("Internal timing\n");
	if(display51(BOARD)) printw("Display on\t\n"); else printw("Display off\n");
	if(field51(BOARD)) printw("Odd  field being Displayed\n"); else printw("Even field being Displayed\n");
	if(vsync51(BOARD)) printw("Vertical sync is     ocurring\n"); else printw("Vertical sync is not ocurring\n");
	if(frameproc51(BOARD)) printw("DT2858 frame processor present\n"); else printw("DT2858 frame processor was not found\n");
	if(exttrig51(BOARD)) printw("External trigger input is high"); else printw("External trigger input is low\t\n");
	refresh();
}

int rand51(int randmax)
{
	return 1+(int)(randmax*(rand()/(RAND_MAX+1.0)));
}


int main(void)
{
	int x;
	int y;
	int i;
	int j;
	int red;
	int green;
	int blue;
	int color;
	int pixel;
	line51 line;
	int fd;
	unsigned char *fgbuf;
	struct sigaction act;
	struct sigaction oldact;

	initscr();
	cbreak();
	erase();
	printw("DT2851 Library Demonstration Program - Unix version\n");
	printw("Copyright (C) 1992...1997 Walter Fetter Lages <w.fetter@ieee.org>.\n\n");
	printw("DT2851 library version: %01d.%02d\n",dtlibver51(),dtlibsubver51());
	printw("Initializing DT2851...\n");
	refresh();
	init51();
	pause51();
	
	erase();
	pausemsg();
	while(!kbhit())	showstatus();
	getch();
	
	erase();
	selectoutbuf51(BOARD,1);
	for(i=0;i < 512;i++) line[i]=255;
	for(i=0;i < 512;i++) writeline51(BOARD,1,i,line);
	setdisplay51(BOARD,1);
	printw("Displaying white\n");
	pausemsg();
	while(!kbhit())	showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
	for(i=0;i < 512;i++) line[i]=0;
	for(i=0;i < 512;i++) writeline51(BOARD,1,i,line);
	setdisplay51(BOARD,1);
	printw("Displaying black\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
	selectoutbuf51(BOARD,0);
	for(i=0;i < 512;i++) line[i]=255;
	for(i=0;i < 512;i++) writeline51(BOARD,0,i,line);
	setdisplay51(BOARD,1);
	printw("Displaying white\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
	for(i=0;i < 256;i++)
	{
		readline51(BOARD,0,i,line);
		for(j=0;j < 512;j++) line[j]=~line[j];
		writeline51(BOARD,0,i+256,line);
	}
	setdisplay51(BOARD,1);
	printw("Displaying half black & half white\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
	for(i=0;i < 512;i++) line[i]=i/2;
	for(i=0;i < 512;i++) writeline51(BOARD,0,i,line);
	setdisplay51(BOARD,1);
	printw("Displaying gray-scale\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
	selectoutlut51(BOARD,1);
	setdisplay51(BOARD,1);
	printw("Displaying inverse gray-scale\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
        selectoutlut51(BOARD,0);
	for(i=0;i < 256;i++) setoutlut51(BOARD,0,i,255*(int)(i/128),255*(int)(i/128),255*(int)(i/128));
	setdisplay51(BOARD,1);
	printw("Displaying half black & half white\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
	for(i=0;i < 256;i++)
	{
		getoutlut51(BOARD,1,i,&red,&green,&blue);
		setoutlut51(BOARD,0,255-i,red,green,blue);
	}
	setdisplay51(BOARD,1);
	printw("Displaying gray-scale\n");
	pausemsg();
	while(!kbhit())	showstatus();
	getch();
	
	erase();
	selectoutlut51(BOARD,0);
	setcursorpos51(BOARD,256,256);
	setcursor51(BOARD,1);
	printw("Cursor at screen center\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setcursor51(BOARD,0);
	setdisplay51(BOARD,0);
	setprotect51(BOARD,PLANE_7_6_5_4_51);
	for(i=0;i < 512;i++) line[i]=i;
	for(i=0;i < 512;i++) writeline51(BOARD,0,i,line);
	setdisplay51(BOARD,1);
	printw("Displaying 16-level gray-scale\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setprotect51(BOARD,NO_PLANE51);
	clearboard51(BOARD);
	printw("Displaying random pixels\n");
	pausemsg();
	selectoutbuf51(BOARD,0);
	for(i=0;i < 512;i++) line[i]=255;
	for(i=0;i < 512;i++) writeline51(BOARD,0,i,line);
	setdisplay51(BOARD,1);
	while(!kbhit())
	{
		x=rand51(512);
		y=rand51(512);
		color=rand51(256);
		pixel=readpixel51(BOARD,0,x,y);
		writepixel51(BOARD,0,x,y,xor(color,pixel));
		showstatus();

	}
	getch();
	
	erase();
	clearboard51(BOARD);
	selectoutbuf51(BOARD,0);
	selectinbuf51(BOARD,0);
	settiming51(BOARD,EXTTIM51);
	setmode51(BOARD,VIDEO_IN51);
	start51(BOARD,MULTIPLE51);
	setdisplay51(BOARD,1);
	printw("Displaying real time camera image\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	stop51(BOARD);
	printw("Displaying frozzen camera image\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
	selectoutbuf51(BOARD,1);
	selectinbuf51(BOARD,1);
	selectinlut51(BOARD,1);
	start51(BOARD,MULTIPLE51);
	setdisplay51(BOARD,1);
	printw("Displaying negative real time camera image\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	stop51(BOARD);
	printw("Displaying frozzen negative camera image\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	setdisplay51(BOARD,0);
	for(i=0;i < 256;i++) setinlut51(BOARD,1,i,getinlut51(BOARD,0,i));
	start51(BOARD,MULTIPLE51);
	setdisplay51(BOARD,1);
	printw("Displaying real time camera image\n");
	pausemsg();
	while(!kbhit()) showstatus();
	getch();
	
	erase();
	stop51(BOARD);
	setdisplay51(BOARD,0);
	selectoutbuf51(BOARD,0);
	selectinbuf51(BOARD,0);
	selectinlut51(BOARD,0);
	act.sa_handler=inttest;
       	act.sa_mask=0;
        act.sa_flags=SA_RESTART;
	sigaction(SIGIO,&act,&oldact);
	enableint51(BOARD,ON_DONE51 | ON_SYNC51);
	printw("Displaying real time camera image\n");
	start51(BOARD,SINGLE51);
	setdisplay51(BOARD,1);
	pausemsg();
	while(!kbhit())
	{
		showstatus();
		printw("\nUsing interrupts\n");
		if(doneflag)
		{
			printw("Done interrupt is ocurring\t\n");
			doneflag=0;
		}
		else printw("Done interrupt is not ocurring\n");
		if(syncflag)
		{
			printw("Sync interrupt is ocurring\t\n");
			syncflag=0;
		}
		else printw("Sync interrupt is not ocurring\n");
		refresh();
	}
	getch();
	stop51(BOARD);
	disableint51(BOARD,ON_DONE51 | ON_SYNC51);
	sigaction(SIGIO,&oldact,NULL);

	erase();
	clearboard51(BOARD);
	selectoutbuf51(BOARD,0);
	fd=open("/dev/fgabuf0",O_RDWR);
      	fgbuf=mmap(0,512*512,PROT_READ | PROT_WRITE,MAP_SHARED,fd,0);
	close(fd);      	
	for(i=0;i < 512*512;i++) fgbuf[i]=0xff;
	setdisplay51(BOARD,1);
	printw("Displaying white\n");
	pausemsg();
	while(!kbhit())	showstatus();
	getch();

	erase();
	setdisplay51(BOARD,0);
	for(i=0;i < 512*512;i++) fgbuf[i]=0x00;
	setdisplay51(BOARD,1);
	printw("Displaying black\n");
	pausemsg();
	while(!kbhit())	showstatus();
	getch();

	erase();
	setdisplay51(BOARD,0);
	munmap(fgbuf,512*512);
	pausemsg();
	while(!kbhit()) showstatus();
	getch();

	erase();
	printw("Releasing DT2851...\n");
	refresh();
	close51();
	pause51();
	
	erase();
	refresh();
	endwin();
	
	return 0;
}

