/****************************************************************************
                                                                   
				DT2851 DRIVER                              
	Copyright (C) 1992...1997 Walter Fetter Lages <w.fetter@ieee.org> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
                                                                   
*****************************************************************************/

#define MODULE

#include <linux/module.h>

extern "C"
{
#include <asm/irq.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/sched.h>
};

#include <asm/io.h>
#include <asm/segment.h>
#include <linux/errno.h>
#include <unistd.h>

#include <dt2851io.h>


// just because I like it this way... 
#define outportw(port,data)	outw(data,port)
#define inportw(port)		inw(port)


// support old naming sillyness 
#if LINUX_VERSION_CODE < 0x020100                                     
#define ioremap vremap
#define iounmap vfree                                                     
#endif

/* don't use #define. This way these paramenters are tunable:
   insmod dt2851 <paramenter>=<value> */

static int major=26;
static unsigned int baseioa=0x230;
static unsigned long basemema=0xa00000;
static int irqa=15;
static unsigned int baseiob=0x230;
static unsigned long basememb=0xa00000;
static int irqb=15;
static unsigned int baseioc=0x230;
static unsigned long basememc=0xa00000;
static int irqc=15;
static unsigned int baseiod=0x230;
static unsigned long basememd=0xa00000;
static int irqd=15;

enum MINOR
{
	BUF0_MINOR=	0,
	BUF1_MINOR=	1,
	ILUT0_MINOR=	2,
	ILUT1_MINOR=	3,
	ILUT2_MINOR=	4,
	ILUT3_MINOR=	5,
	ILUT4_MINOR=	6,
	ILUT5_MINOR=	7,
	ILUT6_MINOR=	8,
	ILUT7_MINOR=	9,
	OLUT0_MINOR=	10,
	OLUT1_MINOR=	11,
	OLUT2_MINOR=	12,
	OLUT3_MINOR=	13,
	OLUT4_MINOR=	14,
	OLUT5_MINOR=	15,
	OLUT6_MINOR=	16,
	OLUT7_MINOR=	17,
	CUR_MINOR=	18,
	CS_MINOR=	19
};

const DT_IO_EXTENT=16;
const BUF_SIZE=512*512;
const LUT_SIZE=256;

// DT2851 registers definitions
enum REGISTERS
{
	INCSR1=	0x0000,	// video input control/status register 1
	INCSR2=	0x0002,	// video input control/status register 2
	OUTCSR=	0x0004,	// video output control/status register
	CURSOR=	0x0006,	// cursor register
	INDEX=	0x0008,	// index  register
	INLUT=	0x000a,	// input look-up table entry register
	REDGRN=	0x000c,	// red and green output look-up table register
	BLUE=	0x000e	// blue output look-up table register
};


// video input control/status register 1 definitions
enum CTRLREG1
{
	BUSY=	0x0080,
	DONEINT=0x0040,	// interrupt on done
	EXTTRG=	0x0020,	// external trigger
	FRPROC=	0x0010,	// frame processor
	ENSTOP=	0x0008,	// enable stop
	ISEL=	0x0007,	// input LUT select
	ISEL2=	0x0004,
	ISEL1=	0x0002,
	ISEL0=	0x0001
};

// video input control/status register 2 definitions
enum CTRLREG2
{
	BUFSEL=		0x0080,	// buffer select
	MODE=		0x0070, // mode
	FEEDBACK=	0x0000,
	VIDEOIN=	0x0010,
	LOADLUT=	0x0040,
	SLOWSCAN=	0x0050,
	PORTOUT=	0x0060,
	PORTIN=		0x0070,
	WP=		0x000f,	// write protect
	WP3=		0x0008,
	WP2=		0x0004,
	WP1=		0x0002,
	WP0=		0x0001
};


// video output control/status register definitions
enum VIDREG
{
	VSYNC=		0x8000,	// vertical sync
	SYNCINT=	0x4000,	// interrupt on sync
	FIELD=		0x2000,
	// EXTTRG=	0x1000,	// This is a duplicate of EXTTRG in INCSR1
	// RESERVED=	0x0800,
	// RESERVED=	0x0400,
	// RESERVED=	0x0200,
	// RESERVED=	0x0100,
	DISP=		0x0080,
	CURS=		0x0040,	// cursor
	EXTTIM=		0x0020,	// external timing
	DISBUF=		0x0010,	// diplay buffer select
	OSEL=		0x000f,	// output LUT select
	OSEL3=		0x0008,
	OSEL2=		0x0004,
	OSEL1=		0x0002,
	OSEL0=		0x0001
};


// cursor register definitions
enum CURREG
{
	LINE=	0xff00,	// line address/2
	PIXEL=	0x00ff	// pixel address/2
};

// index register definition
const LUTENTRY=	0x00ff;	// select LUT entry

// input look-up table entry register definition
const LUTDATA=	0x00ff;	// contains LUT data

// red and green output look-up table register definitions
const GREENDATA=	0xff00;	// green LUT data
const REDDATA=		0x00ff;	// red LUT data

// blue output look-up table register definition
const BLUEDATA=		0x00ff;	// blue LUT data


static unsigned int baseio[4];
static unsigned long basemem[4];
static struct
{
	int irq;
	struct task_struct *p;
} irq_data[4];

int getboard(struct inode *inode)
{
	return MINOR(inode->i_rdev)/(CS_MINOR-BUF0_MINOR+1);
}

int getdev(struct inode *inode)
{
	return MINOR(inode->i_rdev) % (CS_MINOR-BUF0_MINOR+1);
}

void irq_handler(int irq,void *dev_id,struct pt_regs *regs)
{
	for(int i=0; i < 4;i++)
	{
		if((irq==irq_data[i].irq) && (irq_data[i].p != NULL)) send_sig(SIGIO,irq_data[i].p,1);
	}
}

static void stop(int board)
{
	unsigned int incsr1=baseio[board]+INCSR1;
	unsigned short old=inportw(incsr1);
	outportw(incsr1,old | ENSTOP);
	while(inportw(incsr1) & BUSY);
}

static void clearilut(int board,int ilut)
{	
	unsigned int base=baseio[board];
	unsigned short incsr1=inportw(base+INCSR1);
	incsr1|=ENSTOP;
	outportw(base+INCSR1,incsr1);
	outportw(base+INCSR1,incsr1 & ~BUSY);
	do
	{
		incsr1=inportw(base+INCSR1);
	}while(incsr1 & BUSY);

	unsigned short incsr2=inportw(base+INCSR2);
	outportw(base+INCSR2,LOADLUT);

	outportw(base+INCSR1,ilut);
	
	for(int i=0;i < LUT_SIZE;i++)
	{
		outportw(base+INDEX,i);

		switch(ilut)
		{
			case 0:
			{
				// gray-scale
				outportw(base+INLUT,i);
				break;
			}
			case 1:
			{
				// invese gray-scale
				outportw(base+INLUT,~i);
				break;
			}
			case 2:
			{	
				// right-justified 4 MSBs
				outportw(base+INLUT,i >> 4);
				break;
			}
			case 3:
			{
				// left-justified 4 MSBs
				outportw(base+INLUT,i & 0xf0);
				break;
			}
			case 4:
			{
				// left-justified 6 MSBs
				outportw(base+INLUT,i & 0xfc);
				break;
			}
			case 5:
			{
				// left-justified 7 MSBs
				outportw(base+INLUT,i & 0xfe);
				break;
			}
			case 6:
			{
				// bi-level threshold
				if(i > 0x80) outportw(base+INLUT,0xff);
				else outportw(base+INLUT,0x00);
				break;
			}
			case 7:
			{
				// inverse bi-level threshod
				if(i > 0x80) outportw(base+INLUT,0x00);
				else outportw(base+INLUT,0xff);
			}
		}
	}
	outportw(base+INCSR2,incsr2);
	outportw(base+INCSR1,incsr1 & ~(DONEINT | ISEL));
}

static void clearolut(int board,int olut)
{	
	unsigned int base=baseio[board];

	unsigned short incsr1=inportw(base+INCSR1);
	incsr1|=ENSTOP;
	outportw(base+INCSR1,incsr1);
	outportw(base+INCSR1,incsr1 & ~BUSY);
	do
	{
		incsr1=inportw(base+INCSR1);
	}while(incsr1 & BUSY);

	unsigned short outcsr=inportw(base+OUTCSR);
	unsigned short incsr2=inportw(base+INCSR2);
	outportw(base+INCSR2,LOADLUT);
	outportw(base+OUTCSR,olut);
	
	for(int i=0;i < LUT_SIZE;i++)
	{
		outportw(base+INDEX,i);

		switch(olut)
		{
			case 0:
			{
				// gray-scale
				outportw(base+BLUE,i);
				outportw(base+REDGRN,i | (i << 8));
				break;
			}
			case 1:
			{
				// inverse gray-scale
				unsigned short o=~i & 0xff;
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 2:
			{
				// right-justified 4-bit image
				unsigned short o=i << 4;
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 3:
			{
				// left-justified 4-bit image
				unsigned short o=i & 0xf0;
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 4:
			{
				// addition of two 4-bit images
				unsigned short o1=i >> 4;
				unsigned short o2=i & 0x0f;
				unsigned short o=(o1+o2);
				if(o > 0xff) o=0xff;	// saturation
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 5:
			{
				// subtraction of two 4-bit images
				unsigned short o1=i >> 4;
				unsigned short o2=i & 0x0f;
				unsigned short o=(o1-o2) & 0xff;
				if(o & 0x10) o=0; // negative saturation
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 6:
			{
				// gray-scale image with color overlay
				switch(i & 0x03)
				{
					case 0x00:
					{
						outportw(base+BLUE,i);
						outportw(base+REDGRN,i | (i << 8));
						break;
					}
					case 0x01:
					{	
						outportw(base+BLUE,0);
						outportw(base+REDGRN,REDDATA);
						break;
					}
					case 0x02:
					{
						outportw(base+BLUE,0);
						outportw(base+REDGRN,GREENDATA);
						break;
					}
					default:
					{
						outportw(base+BLUE,0xff);
						outportw(base+REDGRN,0x0000);
					}
				}
				break;
			}
			case 7:
			{
				//  gray-scale image with red overlay
				if(i & 1)
				{
					outportw(base+BLUE,0);
					outportw(base+REDGRN,REDDATA);
				}
				else
				{
					outportw(base+BLUE,i);
					outportw(base+REDGRN,i | (i << 8));
				}
				break;
			}
		}
	}
	outportw(base+INCSR2,incsr2);
	outportw(base+OUTCSR,outcsr);
	outportw(base+INCSR1,incsr1 & ~(DONEINT | ISEL));
}

static void clear(int board)
{
	unsigned int base=baseio[board];

	unsigned short outcsr=inportw(base+OUTCSR);
	outportw(base+OUTCSR,outcsr & ~(SYNCINT | DISP | CURS | EXTTIM | DISBUF | OSEL));

	unsigned short incsr1=inportw(base+INCSR1);
	incsr1|=ENSTOP;
	outportw(base+INCSR1,incsr1);
	outportw(base+INCSR1,incsr1 & ~BUSY);


// program look-up tables

	do
	{
		incsr1=inportw(base+INCSR1);
	}while(incsr1 & BUSY);

	outcsr=inportw(base+OUTCSR);
	unsigned short incsr2=inportw(base+INCSR2);
	outportw(base+INCSR2,LOADLUT);
	
	for(int i=0;i < LUT_SIZE;i++)
	{
		unsigned short ilut=inportw(base+INCSR1);
		ilut&=~ISEL;
		unsigned short olut=inportw(base+OUTCSR);
		olut&=~OSEL;
		outportw(base+INDEX,i);

// program input look-up table 0 with gray-scale
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i);

// program output look-up table 0 with gray-scale
		outportw(base+OUTCSR,olut++);
		outportw(base+BLUE,i);
		outportw(base+REDGRN,i | (i << 8));

// program input look-up table 1 with invese gray-scale
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,~i);

// program output look-up table 1 with inverse gray-scale
		outportw(base+OUTCSR,olut++);
		unsigned short o=~i & 0xff;
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));
	
// program input look-up table 2 with right-justified 4 MSBs
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i >> 4);

// program output look-up table 2 to display a right-justified 4-bit image
		outportw(base+OUTCSR,olut++);
		o=i << 4;
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));
		
// program input look-up table 3 with left-justified 4 MSBs
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i & 0xf0);

// program output look-up table 3 to display a left-justified 4-bit image
		outportw(base+OUTCSR,olut++);
		o=i & 0xf0;
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));

// program input look-up table 4 with left-justified 6 MSBs
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i & 0xfc);

// program output look-up table 4 for addition of two 4-bit images
		outportw(base+OUTCSR,olut++);
		unsigned short o1=i >> 4;
		unsigned short o2=i & 0x0f;
		o=(o1+o2);
		if(o > 0xff) o=0xff;	// saturation
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));
		
// program input look-up table 5 with left-justified 7 MSBs
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i & 0xfe);

// program output look-up table 5 for subtraction of two 4-bit images
		outportw(base+OUTCSR,olut++);
		o1=i >> 4;
		o2=i & 0x0f;
		o=(o1-o2) & 0xff;
		if(o & 0x10) o=0; // negative saturation
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));
		

// program input look-up table 6 with bi-level threshold
		outportw(base+INCSR1,ilut++);
		if(i > 0x80) outportw(base+INLUT,0xff);
		else outportw(base+INLUT,0x00);

// program output look-up table 6 for display a gray-scale image with color overlay
		outportw(base+OUTCSR,olut++);
		switch(i & 0x03)
		{
			case 0x00:
			{
				outportw(base+BLUE,i);
				outportw(base+REDGRN,i | (i << 8));
				break;
			}
			case 0x01:
			{	
				outportw(base+BLUE,0);
				outportw(base+REDGRN,REDDATA);
				break;
			}
			case 0x02:
			{
				outportw(base+BLUE,0);
				outportw(base+REDGRN,GREENDATA);
				break;
			}
			default:
			{
				outportw(base+BLUE,0xff);
				outportw(base+REDGRN,0x0000);
			}
		}
		
// program input look-up table 7 with inverse bi-level threshod
		outportw(base+INCSR1,ilut++);
		if(i > 0x80) outportw(base+INLUT,0x00);
		else outportw(base+INLUT,0xff);

// program output look-up table 7 for display a gray-scale image with red overlay
		outportw(base+OUTCSR,olut++);
		if(i & 1)
		{
			outportw(base+BLUE,0);
			outportw(base+REDGRN,REDDATA);
		}
		else
		{
			outportw(base+BLUE,i);
			outportw(base+REDGRN,i | (i << 8));
		}
	}

	outportw(base+INCSR2,incsr2);
	outportw(base+OUTCSR,outcsr);
	outportw(base+INCSR1,incsr1 & ~(DONEINT | ISEL));

	outportw(base+INCSR2,~(BUFSEL | MODE | WP));
	outportw(base+CURSOR,0x8080);


// initialize board memory
	char *p=(char *)ioremap(basemem[board],2*BUF_SIZE);
	if(p != NULL) memset_io(p,(char)0,2*BUF_SIZE);
	iounmap(p);
}

static int buf_lseek(struct inode *inode,struct file *file,off_t offset,int orig)
{
	long npos;

	switch(orig)
	{
		case SEEK_SET:
		{
			npos=offset;
			break;
		}
		case SEEK_CUR:
		{
			npos=file->f_pos+offset;
			break;
		}
		case SEEK_END:
		{
			npos=BUF_SIZE-1-offset;
			break;
		}
		default: return -EINVAL;
	}
	if((npos < 0) || (npos >= BUF_SIZE)) return -EINVAL;
	file->f_pos=npos;
	return file->f_pos;
}

static int buf_read(struct inode *inode,struct file *file,const char *buf,int count)
{
	if(count < 0) return -EINVAL;

	int nbytes=(file->f_pos + count < BUF_SIZE)? count:BUF_SIZE-file->f_pos;

	unsigned char *bufptr=(unsigned char *)file->private_data;

	memcpy_tofs((char *)buf,bufptr+file->f_pos,nbytes);

	file->f_pos+=nbytes;
	return nbytes;
}

static int buf_write(struct inode *inode,struct file *file,const char *buf,int count)
{
	if(count < 0) return -EINVAL;

	int nbytes=(file->f_pos + count < BUF_SIZE)? count:BUF_SIZE-file->f_pos;

	unsigned char *bufptr=(unsigned char *)file->private_data;
	memcpy_fromfs(bufptr+file->f_pos,buf,nbytes);

	file->f_pos+=nbytes;
	return nbytes;
}

static int buf_mmap(struct inode *inode,struct file *file,struct vm_area_struct *vma)
{
	if((vma->vm_end-vma->vm_start+vma->vm_offset) > BUF_SIZE) return -EINVAL;

	int board=getboard(inode);
	int dev=getdev(inode);

	vma->vm_offset+=basemem[board]+(dev-BUF0_MINOR)*BUF_SIZE;

	if(vma->vm_offset & ~PAGE_MASK) return -ENXIO;

	if(remap_page_range(vma->vm_start,vma->vm_offset,vma->vm_end-vma->vm_start,vma->vm_page_prot)) return -EAGAIN;

	vma->vm_inode=inode;
	inode->i_count++;
	return 0;
}

static int buf_open(struct inode *inode,struct file *file)
{
	MOD_INC_USE_COUNT;

	int board=getboard(inode);
	int dev=getdev(inode);

	unsigned char *bufptr=(unsigned char *)ioremap(basemem[board]+(dev-BUF0_MINOR)*BUF_SIZE,BUF_SIZE);
	if(bufptr == NULL)
	{
		MOD_DEC_USE_COUNT;
		return -ENOMEM;
	}
	memset_io(bufptr,(char)0,BUF_SIZE);
	file->private_data=bufptr;
	file->f_pos=0;

	return 0;   
}

static void buf_close(struct inode *inode,struct file *file)
{
	unsigned char *bufptr=(unsigned char *)file->private_data;
	iounmap(bufptr);

	MOD_DEC_USE_COUNT;
}

static struct file_operations buf_fops=
{
	buf_lseek,
	buf_read,	
	buf_write,
	NULL,		/* buf_readdir */
	NULL,		/* buf_select */
	NULL,		/* buf_ioctl */
	buf_mmap,
	buf_open,
	buf_close,
	NULL,		/* buf_fsync */
	NULL,		/* buf_fasync */
	NULL,		/* buf_check_media_change */
	NULL		/* buf_revalidate */
};

static int ilut_lseek(struct inode *inode,struct file *file,off_t offset,int orig)
{
	long npos;

	switch(orig)
	{
		case SEEK_SET:
		{
			npos=offset;
			break;
		}
		case SEEK_CUR:
		{
			npos=file->f_pos+offset;
			break;
		}
		case SEEK_END:
		{
			npos=BUF_SIZE-1-offset;
			break;
		}
		default: return -EINVAL;
	}
	if((npos < 0) || (npos >= LUT_SIZE)) return -EINVAL;
	file->f_pos=npos;
	return file->f_pos;
}

static int ilut_read(struct inode *inode,struct file *file,const char *buf,int count)
{
	if(count < 0) return -EINVAL;

	int nbytes=(file->f_pos + count < LUT_SIZE)? count:LUT_SIZE-file->f_pos;

	int board=getboard(inode);
	int lut=getdev(inode)-ILUT0_MINOR;
	
	int base=baseio[board];

	stop(board);
	unsigned short old2=inportw(base+INCSR2);
	outportw(base+INCSR2,(old2 & ~MODE) | LOADLUT);
	unsigned short old1=inportw(base+INCSR1);
	outportw(base+INCSR1,(old1 & ~ISEL) | (lut & ISEL));
	
	for(int i=0;i < nbytes;i++)
	{
		outportw(base+INDEX,file->f_pos++);
		unsigned char lutdata=inportw(base+INLUT);
		put_user(lutdata,(unsigned char *)buf++);
	}
	
	outportw(base+INCSR2,old2);
	outportw(base+INCSR1,old1);	

	return nbytes;
}

static int ilut_write(struct inode *inode,struct file *file,const char *buf,int count)
{
	if(count < 0) return -EINVAL;

	int nbytes=(file->f_pos + count < LUT_SIZE)? count:LUT_SIZE-file->f_pos;

	int board=getboard(inode);
	int lut=getdev(inode)-ILUT0_MINOR;
	
	int base=baseio[board];

	stop(board);
	unsigned short old2=inportw(base+INCSR2);
	outportw(base+INCSR2,(old2 & ~MODE) | LOADLUT);
	unsigned short old1=inportw(base+INCSR1);
	outportw(base+INCSR1,(old1 & ~ISEL) | (lut & ISEL));
	
	for(int i=0;i < nbytes;i++)
	{
		outportw(base+INDEX,file->f_pos++);
		unsigned char lutdata=get_user(buf++);
		outportw(base+INLUT,lutdata);
	}
	
	outportw(base+INCSR2,old2);
	outportw(base+INCSR1,old1);	

	return nbytes;
}

static int ilut_open(struct inode *inode,struct file *file)
{
	MOD_INC_USE_COUNT;
	int ilut=getdev(inode)-ILUT0_MINOR;
	clearilut(getboard(inode),ilut);

	file->f_pos=0;
	return 0;   
}

static void ilut_close( struct inode *inode,struct file *file)
{
	MOD_DEC_USE_COUNT;
}

static struct file_operations ilut_fops=
{
	ilut_lseek,
	ilut_read,	
	ilut_write,
	NULL,		/* ilut_readdir */
	NULL,		/* ilut_select */
	NULL,		/* ilut_ioctl */
	NULL,		/* ilut_mmap */
	ilut_open,
	ilut_close,
	NULL,		/* ilut_fsync */
	NULL,		/* ilut_fasync */
	NULL,		/* ilut_check_media_change */
	NULL		/* ilut_revalidate */
};

static int olut_lseek(struct inode *inode,struct file *file,off_t offset,int orig)
{
	long npos;

	switch(orig)
	{
		case SEEK_SET:
		{
			npos=offset;
			break;
		}
		case SEEK_CUR:
		{
			npos=file->f_pos+offset;
			break;
		}
		case SEEK_END:
		{
			npos=BUF_SIZE-1-offset;
			break;
		}
		default: return -EINVAL;
	}
	if((npos < 0) || (npos >= LUT_SIZE)) return -EINVAL;
	file->f_pos=npos;
	return file->f_pos;
}

static int olut_read(struct inode *inode,struct file *file,const char *buf,int count)
{
	if((count < 0) || (count % 3)) return -EINVAL;

	int nbytes=(file->f_pos + count/3 < LUT_SIZE)? count:3*(LUT_SIZE-file->f_pos);

	int board=getboard(inode);
	int lut=getdev(inode)-OLUT0_MINOR;
	
	int base=baseio[board];

	stop(board);
	unsigned short old2=inportw(base+INCSR2);
	outportw(base+INCSR2,(old2 & ~MODE) | LOADLUT);
	unsigned short old=inportw(base+OUTCSR);
	outportw(base+OUTCSR,(old & ~OSEL) | (lut & OSEL));
	for(int i=0;i < nbytes/3;i++)
	{
		outportw(base+INDEX,file->f_pos++);
		unsigned short lutdata=inportw(base+REDGRN);
		put_user(lutdata,(unsigned short *) buf);
		buf+=2;
		lutdata=inportw(base+BLUE);
		put_user((unsigned char)lutdata,(unsigned char *)buf++);
	}
	
	outportw(base+INCSR2,old2);
	outportw(base+OUTCSR,old);	

	return nbytes;
}

static int olut_write(struct inode *inode,struct file *file,const char *buf,int count)
{
	if((count < 0) || (count % 3)) return -EINVAL;

	int nbytes=(file->f_pos + count/3 < LUT_SIZE)? count:3*(LUT_SIZE-file->f_pos);

	int board=getboard(inode);
	int lut=getdev(inode)-OLUT0_MINOR;
	
	int base=baseio[board];

	stop(board);
	unsigned short old2=inportw(base+INCSR2);
	outportw(base+INCSR2,(old2 & ~MODE) | LOADLUT);
	unsigned short old=inportw(base+OUTCSR);
	outportw(base+OUTCSR,(old & ~OSEL) | (lut & OSEL));
	
	for(int i=0;i < nbytes/3;i++)
	{
		outportw(base+INDEX,file->f_pos++);
		
		unsigned short lutdata=get_user((unsigned short *) buf);
		outportw(base+REDGRN,lutdata);
		buf+=2;
		lutdata=get_user((unsigned short *) buf);
		outportw(base+BLUE,lutdata);
	}
	
	outportw(base+INCSR2,old2);
	outportw(base+OUTCSR,old);	

	return nbytes;
}

static int olut_open(struct inode *inode,struct file *file)
{
	MOD_INC_USE_COUNT;
	int olut=getdev(inode)-OLUT0_MINOR;
	clearolut(getboard(inode),olut);
	file->f_pos=0;
	return 0;   
}

static void olut_close( struct inode *inode,struct file *file)
{
	MOD_DEC_USE_COUNT;
}

static struct file_operations olut_fops=
{
	olut_lseek,
	olut_read,	
	olut_write,
	NULL,		/* olut_readdir */
	NULL,		/* olut_select */
	NULL,		/* olut_ioctl */
	NULL,		/* olut_mmap */
	olut_open,
	olut_close,
	NULL,		/* olut_fsync */
	NULL,		/* olut_fasync */
	NULL,		/* olut_check_media_change */
	NULL		/* olut_revalidate */
};

static int cur_lseek(struct inode *inode,struct file *file,off_t offset,int orig)
{
	long npos;

	switch(orig)
	{
		case SEEK_SET:
		{
			npos=offset;
			break;
		}
		case SEEK_CUR:
		{
			npos=file->f_pos+offset;
			break;
		}
		case SEEK_END:
		{
			npos=BUF_SIZE-1-offset;
			break;
		}
		default: return -EINVAL;
	}
	if((npos < 0) || (npos >= BUF_SIZE)) return -EINVAL;
	file->f_pos=npos;
	
	unsigned short x=npos%512;
	unsigned short y=npos/512;
	
	unsigned short y_2_x_2=((y/2) << 8) | (x/2);

	int board=getboard(inode);
	outportw(baseio[board]+CURSOR,y_2_x_2);

	return file->f_pos;
}

static int cur_read(struct inode *inode,struct file *file,const char *buf,int count)
{
	if(count != 2*sizeof(int)) return -EINVAL;

	int board=getboard(inode);
	unsigned short cursor=inportw(baseio[board]+CURSOR);
	int x=(cursor & 0xff)*2;
	put_user(x,(int *)buf);
	int y=(cursor >> 8)*2;
	put_user(y,(int *)(buf+sizeof(int)));
	return 2*sizeof(int);
}

static int cur_write(struct inode *inode,struct file *file,const char *buf,int count)
{
	if(count != 2*sizeof(int)) return -EINVAL;

	unsigned short x=get_user((int *)buf);
	unsigned short y=get_user((int *)(buf+sizeof(int)));
	
	unsigned short y_2_x_2=((y/2) << 8) | (x/2);

	int board=getboard(inode);
	outportw(baseio[board]+CURSOR,y_2_x_2);

	return 2*sizeof(int);
}

static int cur_ioctl(struct inode *inode,struct file *file,unsigned int cmd,unsigned long arg)
{
	int board=getboard(inode);
	unsigned int curport=baseio[board]+OUTCSR;
	unsigned short old=inportw(curport);
	switch(cmd)
	{
		case CUR_ON:
		{
			outportw(curport,old | CURS);
			break;
		}
		case CUR_OFF:
		{
			outportw(curport,old & ~CURS);
			break;
		}
		case CUR_QUERY:	
		{
			int status=(inportw(curport) & CURS) >> 6;
			put_user(status,(int *)arg);
			break;
		}
		default: return -EINVAL;
	}
	return 0;
}

static int cur_open(struct inode *inode,struct file *file)
{
	MOD_INC_USE_COUNT;
	file->f_pos=BUF_SIZE/2;
	int board=getboard(inode);
	unsigned int base=baseio[board];
	unsigned short old=inportw(base+OUTCSR);
	outportw(base+OUTCSR,old & ~CURS);
	outportw(base+CURSOR,0x8080);

	return 0;   
}

static void cur_close( struct inode *inode,struct file *file)
{
	int board=getboard(inode);
	unsigned int curport=baseio[board]+OUTCSR;
	unsigned short old=inportw(curport);
	outportw(curport,old & ~CURS);

	MOD_DEC_USE_COUNT;
}

static struct file_operations cur_fops=
{
	cur_lseek,
	cur_read,	
	cur_write,
	NULL,		/* cur_readdir */
	NULL,		/* cur_select */
	cur_ioctl,
	NULL,		/* cur_mmap */
	cur_open,
	cur_close,
	NULL,		/* cur_fsync */
	NULL,		/* cur_fasync */
	NULL,		/* cur_check_media_change */
	NULL		/* cur_revalidate */
};

static int cs_lseek(struct inode *inode,struct file *file,off_t offset,int orig)
{
	return -EINVAL;
}

static int cs_ioctl(struct inode *inode,struct file *file,unsigned int cmd,unsigned long arg)
{
	int board=getboard(inode);
	unsigned int base=baseio[board];
	switch(cmd)
	{
		case CS_CLEAR:
		{
			clear(board);
			break;
		}
		case CS_SETMODE:
		{
			stop(board);
			int mode=(arg << 4) & MODE;
			unsigned short incsr2=inportw(base+INCSR2);
			outportw(base+INCSR2,(incsr2 & ~MODE) | mode);
			break;
		}
		case CS_MODE:
		{
			int mode=(inportw(base+INCSR2) & MODE) >> 4;
			put_user(mode,(int *)arg);
			break;
		}
		case CS_ENABLEINT:
		{
			if(!request_irq(irq_data[board].irq,irq_handler,0,"DT2851",NULL))
			{
				irq_data[board].p=current;
				enable_irq(irq_data[board].irq);

				unsigned short incsr1=inportw(base+INCSR1);
				int doneint=(arg << 6) & DONEINT;
				outportw(base+INCSR1,(incsr1 & ~DONEINT) | doneint);

				unsigned short outcsr=inportw(base+OUTCSR);
				int syncint=(arg << (14-1)) & SYNCINT;
				outportw(base+OUTCSR,(outcsr & ~SYNCINT) | syncint);
			}
			break;
		}
		case CS_DISABLEINT:
		{
			unsigned short incsr1=inportw(base+INCSR1);
			int doneint=(arg << 6) & DONEINT;
			outportw(base+INCSR1,(incsr1 & ~DONEINT) & ~doneint);

			unsigned short outcsr=inportw(base+OUTCSR);
			int syncint=(arg << (14-1)) & SYNCINT;
			outportw(base+OUTCSR,(outcsr & ~SYNCINT) & ~syncint);

			disable_irq(irq_data[board].irq);
			irq_data[board].p=NULL;
			free_irq(irq_data[board].irq,NULL);
			break;
		}
		case CS_INTACK:
		{
			unsigned short incsr1=inportw(base+INCSR1);
			outportw(base+INCSR1,(incsr1 & ~DONEINT));
			unsigned short outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,(outcsr & ~SYNCINT));

			outportw(base+INCSR1,incsr1);	
			outportw(base+OUTCSR,outcsr);
			break;
		}
		case CS_SELILUT:
		{
			int lut=arg & ISEL;
			unsigned short incsr1=inportw(base+INCSR1);
			outportw(base+INCSR1,(incsr1 & ~ISEL) | lut);
			break;
		}
		case CS_ILUT:
		{
			int lut=inportw(base+INCSR1) & ISEL;
			put_user(lut,(int *)arg);
			break;
		}
		case CS_SELOLUT:
		{
			int lut=arg & OSEL;
			unsigned short outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,(outcsr & ~OSEL) | lut);
			break;
		}
		case CS_OLUT:
		{
			int lut=inportw(base+OUTCSR) & OSEL;
			put_user(lut,(int *)arg);
			break;
		}
		case CS_SELIBUF:
		{
			stop(board);
			int buf=(arg << 7) & BUFSEL;
			unsigned short incsr2=inportw(base+INCSR2);
			outportw(base+INCSR2,(incsr2 & ~BUFSEL) | buf);
			break;
		}
		case CS_IBUF:
		{
			int buf=(inportw(base+INCSR2) & BUFSEL) >> 7;
			put_user(buf,(int *)arg);
			break;
		}
		case CS_SELOBUF:
		{
			int buf=(arg << 4) & DISBUF;
			unsigned short outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,(outcsr & ~DISBUF) | buf);
			break;
		}
		case CS_OBUF:
		{
			int buf=(inportw(base+OUTCSR) & DISBUF) >> 4;
			put_user(buf,(int *)arg);
			break;
		}
		case CS_SETPROT:
		{
			stop(board);
			int prot=arg & WP;
			unsigned short incsr2=inportw(base+INCSR2);
			outportw(base+INCSR2,(incsr2 & ~WP) | prot);
			break;
		}
		case CS_PROT:
		{
			int buf=inportw(base+INCSR2) & WP;
			put_user(buf,(int *)arg);
			break;
		}
		case CS_START:
		{
			stop(board);
			int opmode=((~arg) << 3) & ENSTOP;
			unsigned short incsr1=inportw(base+INCSR1);
			outportw(base+INCSR1,(incsr1 & ~ENSTOP) | opmode);
			outportw(base+INCSR1,(incsr1 & ~ENSTOP) | opmode | BUSY);
			break;
		}
		case CS_STOP:
		{
			stop(board);
			break;
		}
		case CS_BUSY:
		{
			int busy=(inportw(base+INCSR1) & BUSY) >> 7;
			put_user(busy,(int *)arg);
			break;
		}
		case CS_DISPLAYON:
		{
			unsigned short outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,outcsr | DISP);
			break;
		}
		case CS_DISPLAYOFF:
		{
			unsigned short outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,outcsr & ~DISP);
			break;
		}
		case CS_SETTIMING:
		{
			int timing=(arg << 5) & EXTTIM;
			unsigned short outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,(outcsr & ~EXTTIM) | timing);
			break;
		}
		case CS_TIMING:
		{
			int timing=(inportw(base+OUTCSR) & EXTTIM) >> 5;
			put_user(timing,(int *)arg);
			break;
		}
		case CS_DISPLAY:
		{
			int display=(inportw(base+OUTCSR) & DISP) >> 7;
			put_user(display,(int *)arg);
			break;
		}
		case CS_FIELD:
		{
			int field=(inportw(base+OUTCSR) & FIELD) >> 13;
			put_user(field,(int *)arg);
			break;
		}
		case CS_VSYNC:
		{
			int vsync=(inportw(base+OUTCSR) & VSYNC) >> 15;
			put_user(vsync,(int *)arg);
			break;
		}
		case CS_FRPROC:
		{
			int frproc=(~inportw(base+INCSR1) & FRPROC) >> 4;
			put_user(frproc,(int *)arg);
			break;
		}
		case CS_EXTTRG:
		{
			int exttrg=(inportw(base+INCSR1) & EXTTRG) >> 5;
			put_user(exttrg,(int *)arg);
			break;
		}

		default: return -EINVAL;
	}
	return 0;
}

static int cs_open(struct inode *inode,struct file *file)
{
	MOD_INC_USE_COUNT;

	int board=getboard(inode);
	unsigned int base=baseio[board];

	unsigned short outcsr=inportw(base+OUTCSR);
	outportw(base+OUTCSR,outcsr & ~(SYNCINT | DISP | CURS | EXTTIM | DISBUF | OSEL));

	unsigned short incsr1=inportw(base+INCSR1);
	incsr1|=ENSTOP;
	outportw(base+INCSR1,incsr1);
	outportw(base+INCSR1,incsr1 & ~BUSY);

	outportw(base+INCSR2,~(BUFSEL | MODE | WP));

	return 0;   

}

static void cs_close( struct inode *inode,struct file *file)
{
	MOD_DEC_USE_COUNT;
}

static struct file_operations cs_fops=
{
	cs_lseek,
	NULL,		/* cs_read */	
	NULL,		/* cs_write */
	NULL,		/* cs_readdir */
	NULL,		/* cs_select */
	cs_ioctl,
	NULL,		/* cs_mmap */
	cs_open,
	cs_close,
	NULL,		/* cs_fsync */
	NULL,		/* cs_fasync */
	NULL,		/* cs_check_media_change */
	NULL		/* cs_revalidate */
};


static int dt_open(struct inode *inode,struct file *file)
{
	int minor=MINOR(inode->i_rdev);
	for(int i=1;i < 4;i++)
	{
		if(minor > CS_MINOR) minor-=CS_MINOR+1;
	}

	if((minor >= BUF0_MINOR) && (minor <= BUF1_MINOR)) file->f_op= &buf_fops;
	else if((minor >= ILUT0_MINOR) && (minor <= ILUT7_MINOR)) file->f_op = &ilut_fops;
	else if((minor >= OLUT0_MINOR) && (minor <= OLUT7_MINOR)) file->f_op = &olut_fops;
	else if(minor == CUR_MINOR) file->f_op=&cur_fops;
	else if(minor == CS_MINOR) file->f_op=&cs_fops;
	else return -ENXIO;

	if(file->f_op && file->f_op->open) return file->f_op->open(inode,file);
	return 0;   
}


static struct file_operations dt_fops=
{
	NULL,		/* dt_lseek */
	NULL,		/* dt_read */	
	NULL,		/* dt_write */
	NULL,		/* dt_readdir */
	NULL,		/* dt_select */
	NULL,		/* dt_ioctl */
	NULL,		/* dt_mmap */
	dt_open,	/* just a selector for the real open */
	NULL,		/* dt_close */
	NULL,		/* dt_fsync */
	NULL,		/* dt_fasync */
	NULL,		/* dt_check_media_change */
	NULL		/* rov_revalidate */
};

extern "C" int init_module(void)
{
	printk("DT2851: ");
	printk("Copyright (C) 1997 Walter Fetter Lages <w.fetter@ieee.org>.\n");
	
	baseio[0]=baseioa;
	basemem[0]=basemema;
	irq_data[0].irq=irqa;
	irq_data[0].p=NULL;

	baseio[1]=baseiob;
	basemem[1]=basememb;
	irq_data[1].irq=irqb;
	irq_data[1].p=NULL;

	baseio[2]=baseioc;
	basemem[2]=basememc;
	irq_data[2].irq=irqc;
	irq_data[2].p=NULL;

	baseio[3]=baseiod;
	basemem[3]=basememd;
	irq_data[3].irq=irqd;
	irq_data[3].p=NULL;

	if(register_chrdev(major,"DT2851",&dt_fops))
	{
		printk("major %d can't be registered.\n",major);
		return -EIO;
	}

	int err=0;

	if(check_region(baseio[0],DT_IO_EXTENT))
	{
		printk("I/O ports are already allocated to another device.\n");
		err=-EIO;
	}
	request_region(baseio[0],DT_IO_EXTENT,"DT2851");

	for(int i=1;i < 4;i++)
	{	
		if(baseio[i]!=baseio[i-1])
		{
			if(check_region(baseio[i],DT_IO_EXTENT))
			{
				printk("I/O ports are already allocated to another device.\n");
				err=-EIO;
			}
			request_region(baseio[i],DT_IO_EXTENT,"DT2851");
		}
	}
	if(err) for(int i=0;i < 4;i++) release_region(baseio[i],DT_IO_EXTENT);

	return err;
}

extern "C" void cleanup_module(void)
{
	for(int i=0;i < 4;i++) release_region(baseio[i],DT_IO_EXTENT);
	if(unregister_chrdev(major,"DT2851") != 0) printk("major %d can't be unregistered\n",major);
}

