/****************************************************************************
                                                                   
				DT2851 LIBRARY                              
	Copyright (C) 1992...1997 Walter Fetter Lages <w.fetter@ieee.org> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
                                                                   
*****************************************************************************/
                                                                   
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>

#include <dt2851.h>
#include <dt2851io.h>

struct boarddev
{
	unsigned char *bufptr[2];
/*	int buf[2]; */
	int ilut[8];
	int olut[8];
	int cur;
	int cs;
};

static struct boarddev boards[4];
	
	
/* dt2851 library/driver version number */
int dtlibver51(void)                      
{
	return DTLIBVERSION;
}

/* dt library/driver sub-version number */
int dtlibsubver51(void)
{
	return DTLIBSUBVERSION;
}

 /* set base address */
void setbaseadd51(int board,unsigned long memadd,unsigned ioadd)
{
/* dummy function on unix, just for compatibility with DOS version */
}

/* open library use */
void init51(void)
{
	int i;
	int j;
	int fd;

	char devname[16]="/dev/fg??????";
	for(i=0;i < 4;i++)
	{
		devname[7]='a'+i;

		devname[8]='b';
		devname[9]='u';
		devname[10]='f';
		devname[12]='\0';
		for(j=0;j < 2;j++)
		{
			devname[11]='0'+j;
/*			boards[i].buf[j]=open(devname,O_RDWR); */
			fd=open(devname,O_RDWR);
			boards[i].bufptr[j]=mmap(0,512*512,PROT_READ | PROT_WRITE,MAP_SHARED,fd,0);
			close(fd);
		}

		devname[8]='i';
		devname[9]='l';
		devname[10]='u';
		devname[11]='t';
		devname[13]='\0';
		for(j=0;j < 8;j++)
		{
			devname[12]='0'+j;
			boards[i].ilut[j]=open(devname,O_RDWR);
		}

		devname[8]='o';
		devname[9]='l';
		devname[10]='u';
		devname[11]='t';
		devname[13]='\0';
		for(j=0;j < 8;j++)
		{
			devname[12]='0'+j;
			boards[i].olut[j]=open(devname,O_RDWR);
		}

		devname[8]='c';
		devname[9]='u';
		devname[10]='r';
		devname[11]='\0';
		boards[i].cur=open(devname,O_RDWR);

		devname[8]='c';
		devname[9]='s';
		devname[10]='\0';
		boards[i].cs=open(devname,O_RDWR);

/*		clearboard51(i); */
	}
}

/* initialize one board */
void clearboard51(int board)
{
	ioctl(boards[board].cs,CS_CLEAR);
}

/* set operation mode */
void setmode51(int board,int mode)
{
	ioctl(boards[board].cs,CS_SETMODE,mode);
}

/* get operation mode */
int getmode51(int board)
{
	int mode;
	ioctl(boards[board].cs,CS_MODE,&mode);
	return mode;
}

/* enable interrupt */
void enableint51(int board,int intr)
{
	ioctl(boards[board].cs,CS_ENABLEINT,intr);
}

/* disable interrupt */
void disableint51(int board,int intr)
{
	ioctl(boards[board].cs,CS_DISABLEINT,intr);
}

/* interrupt acknowledge */
void intack51(int board)
{
	ioctl(boards[board].cs,CS_INTACK);
}

/* select input look-up table */
void selectinlut51(int board,int inlut)
{
	ioctl(boards[board].cs,CS_SELILUT,inlut);
}

/* return selected input look-up  table */
int inlut51(int board)
{
	int lut;
	ioctl(boards[board].cs,CS_ILUT,&lut);
	return lut;
}

/* select output look-up table */
void selectoutlut51(int board,int outlut)
{
	ioctl(boards[board].cs,CS_SELOLUT,outlut);
}

/* return selected output look-up table */
int outlut51(int board)
{
	int lut;
	ioctl(boards[board].cs,CS_OLUT,&lut);
	return lut;
}

/* select input buffer */
void selectinbuf51(int board,int inbuffer)
{
	ioctl(boards[board].cs,CS_SELIBUF,inbuffer);
}

/* return selected input buffer */
int inbuf51(int board)
{
	int buf;
	ioctl(boards[board].cs,CS_ILUT,&buf);
	return buf;
}

/* select output buffer */
void selectoutbuf51(int board,int outbuffer)
{
	ioctl(boards[board].cs,CS_SELOBUF,outbuffer);
}

/* return selected output buffer */
int outbuf51(int board)
{
	int buf;
	ioctl(boards[board].cs,CS_OBUF,&buf);
	return buf;
}

/* set write protection to bit-planes */
void setprotect51(int board,int planes)
{
	ioctl(boards[board].cs,CS_SETPROT,planes);
}

/* return write protected bit-planes */
int protected51(int board)
{
	int planes;
	ioctl(boards[board].cs,CS_PROT,&planes);
	return planes;
}

/* start board operation on single operation or multiple operation mode */
void start51(int board,int opmode)
{
	ioctl(boards[board].cs,CS_START,opmode);
}

/* stop board operation after termination */
void stop51(int board)
{
	ioctl(boards[board].cs,CS_STOP);
}

/* return 1 if an operation is in progress */
int busy51(int board)
{
	int busy;

	ioctl(boards[board].cs,CS_BUSY,&busy);
	return busy;
}

/* turn cursor on and off */
void setcursor51(int board,int status)
{
	if(status) ioctl(boards[board].cur,CUR_ON);
	else ioctl(boards[board].cur,CUR_OFF);
}

/* return cursor status */
int cursor51(int board)
{
	int status;
	ioctl(boards[board].cur,CUR_QUERY,&status);
	return status;
}

/* set  cursor position */
void setcursorpos51(int board,int x,int y)
{
	int buf[2];

	buf[0]=x;
	buf[1]=y;
	write(boards[board].cur,buf,sizeof(buf));
}

/* get cursor position */
void getcursorpos51(int board,int *x,int *y)
{
	int buf[2];

	read(boards[board].cur,buf,sizeof(buf));
	*x=buf[0];
	*y=buf[1];
}

/* turn display on and off */
void setdisplay51(int board,int status)
{
	if(status) ioctl(boards[board].cs,CS_DISPLAYON);
	else ioctl(boards[board].cs,CS_DISPLAYOFF);
}

/* set video timing source */
void settiming51(int board,int source)
{
	ioctl(boards[board].cs,CS_SETTIMING,source);
}

/* return video timing source */
int timing51(int board)
{
	int timing;
	ioctl(boards[board].cs,CS_TIMING,&timing);
	return timing;
}

/* return display status */
int display51(int board)
{
	int display;
	ioctl(boards[board].cs,CS_DISPLAY,&display);
	return display;
}

/* set input lut entry */
void setinlut51(int board,int lut,int index,int value)
{
	lseek(boards[board].ilut[lut],index,SEEK_SET);
	write(boards[board].ilut[lut],&value,sizeof(unsigned char));
}

/* get input lut value */
int getinlut51(int board,int lut,int index)
{
	unsigned char value;

	lseek(boards[board].ilut[lut],index,SEEK_SET);
	read(boards[board].ilut[lut],&value,sizeof(value));
	return value;
}

/* set output lut entry */
void setoutlut51(int board,int lut,int index,int red,int green,int blue)
{
	unsigned char v[3];

	v[0]=red;
	v[1]=green;
	v[2]=blue;
	lseek(boards[board].olut[lut],index,SEEK_SET);
	write(boards[board].olut[lut],&v,sizeof(v));
}

/* get output lut values */
void getoutlut51(int board,int lut,int index,int far *red,int far *green,int far *blue)
{
	unsigned char value[3];

	lseek(boards[board].olut[lut],index,SEEK_SET);
	read(boards[board].olut[lut],&value,sizeof(value));
	*red=value[0];
	*green=value[1];
	*blue=value[0];
}

/* return the field being displayed or digitized */
int field51(int board)
{
	int field;
	ioctl(boards[board].cs,CS_FIELD,&field);
	return field;
}

/* return 1 if vertical sync is ocurring */
int vsync51(int board)
{
	int vsync;
	ioctl(boards[board].cs,CS_VSYNC,&vsync);
	return vsync;
}

/* return 1 if there is a dt2858  board in the system */
int frameproc51(int board)
{
	int frproc;
	ioctl(boards[board].cs,CS_FRPROC,&frproc);
	return frproc;
}

/* return the state of the al trigger input */
int exttrig51(int board)
{
	int exttrg;
	ioctl(boards[board].cs,CS_EXTTRG,&exttrg);
	return exttrg;
}

/* write a picture line */
void writeline51(int board,int buffer,int line,void *lineptr)
{
	int i;
	unsigned char *ptr=boards[board].bufptr[buffer]+line*sizeof(line51);
	for(i=0;i < sizeof(line51);i++)  *(ptr++)=*((unsigned char *)lineptr++);

/*	lseek(boards[board].buf[buffer],line*sizeof(line51),SEEK_SET);
	write(boards[board].buf[buffer],lineptr,sizeof(line51)); */
}

/* read a picture line */
void readline51(int board,int buffer,int line,void *lineptr)
{
	int i;
	unsigned char *ptr=boards[board].bufptr[buffer]+line*sizeof(line51);
	for(i=0;i < sizeof(line51);i++)  *((unsigned char *)lineptr++)=*(ptr++);

/*	lseek(boards[board].buf[buffer],line*sizeof(line51),SEEK_SET);
	read(boards[board].buf[buffer],lineptr,sizeof(line51)); */
}

/* write a pixel */
void writepixel51(int board,int buffer,int x,int y,int color)
{
	unsigned char *ptr=boards[board].bufptr[buffer]+y*sizeof(line51)+x;
	*ptr=color;
/*	lseek(boards[board].buf[buffer],y*sizeof(line51)+x,SEEK_SET);
	write(boards[board].buf[buffer],&color,sizeof(unsigned char)); */
}

/* read a pixel */
int readpixel51(int board,int buffer,int x,int y)
{
	unsigned char *ptr=boards[board].bufptr[buffer]+y*sizeof(line51)+x;
	return *ptr;

/*	unsigned char color;

	lseek(boards[board].buf[buffer],y*sizeof(line51)+x,SEEK_SET);
	read(boards[board].buf[buffer],&color,sizeof(color));
	return color; */
}

/* close library use */
void close51(void)
{
	int i;
	int j;
	
	for(i=0;i < 4;i++)
	{
		for(j=0;j < 2;j++) munmap(boards[i].bufptr[j],512*512);
/* close(boards[i].buf[j]); */
		for(j=0;j < 8;j++) close(boards[i].ilut[j]);
		for(j=0;j < 8;j++) close(boards[i].olut[j]);
		close(boards[i].cur);
		close(boards[i].cs);
	}
}

